
#include "hash.h"

lowerHash *lowerHash_new(int noOfSlot)
{
	lowerHash *returnValue;
	int i;

	sequenceSet *aSequenceSet;

	returnValue = (lowerHash *) malloc(sizeof(lowerHash));
	memset(returnValue, 0, sizeof(lowerHash));

	returnValue->noOfSlot = noOfSlot;
	returnValue->slot = (sequenceSet **) malloc(sizeof(sequenceSet *)*noOfSlot);

	for (i = 0; i < noOfSlot; i++)
	{
		aSequenceSet = sequenceSet_new();
		returnValue->slot[i] = aSequenceSet;
	}

	return returnValue;
}

int lowerHash_slotNo(sequence *aSequence, int noOfSlot)
{
	int returnValue;

	int i;

	int noOfElt;
	int *elt;

	noOfElt = aSequence->noOfElt;
	elt = aSequence->elt;

	returnValue = 0;
	for (i = 0; i < noOfElt; i++)
	{
		returnValue += elt[i];
	}

	returnValue = returnValue%noOfSlot;

	return returnValue;
}


// non-redundanct insertion
void lowerHash_insert(lowerHash *aLowerHash, sequence *aSequence)
{
	int slotNo;
	sequenceSet *aSequenceSet;

	int isExist;

	slotNo = lowerHash_slotNo(aSequence, aLowerHash->noOfSlot);

	aSequenceSet = aLowerHash->slot[slotNo];

	isExist = sequenceSet_isExist(aSequenceSet, aSequence);
	if (isExist == 0)
	{
		sequenceSet_insert(aSequenceSet, aSequence);
	}
}

int lowerHash_isExist(lowerHash *aLowerHash, sequence *aSequence)
{
	int returnValue;

	int noOfSlot;
	sequenceSet **slot;

	int slotNo;

	sequenceSet *aSequenceSet;

	

	noOfSlot = aLowerHash->noOfSlot;
	slot = aLowerHash->slot;

	slotNo = lowerHash_slotNo(aSequence, aLowerHash->noOfSlot);
	aSequenceSet = aLowerHash->slot[slotNo];

	returnValue = sequenceSet_isExist(aSequenceSet, aSequence);

	return returnValue;
}


int upperHash_isExist(upperHash *aUpperHash, sequence *aSequence)
{
	int returnValue;

	int noOfSlot;
	lowerHash **slot;

	lowerHash *aLowerHash;

	int firstElt;


	noOfSlot = aUpperHash->noOfSlot;
	slot = aUpperHash->slot;

	if (aSequence->noOfElt > 0)
	{
		firstElt = aSequence->elt[0];

		aLowerHash = slot[firstElt];

		returnValue = lowerHash_isExist(aLowerHash, aSequence);
	}
	else
	{
		returnValue = 0;
	}

	return returnValue;
}

upperHash *upperHash_new(int upper_noOfSlot, int lower_noOfSlot)
{
	upperHash *returnValue;
	int i;

	lowerHash *aLowerHash;

	returnValue = (upperHash *) malloc(sizeof(upperHash));
	memset(returnValue, 0, sizeof(upperHash));

	returnValue->noOfSlot = upper_noOfSlot;
	returnValue->slot = (lowerHash **) malloc(sizeof(lowerHash *)*upper_noOfSlot);

	for (i = 0; i < upper_noOfSlot; i++)
	{
		aLowerHash = lowerHash_new(lower_noOfSlot);
		returnValue->slot[i] = aLowerHash;
	}

	return returnValue;
}

// assume aSequene of length 2
int upperHash_slotNo(sequence *aSequence)
{
	int returnValue;

	int noOfElt;
	int *elt;

	noOfElt = aSequence->noOfElt;
	elt = aSequence->elt;

	returnValue = elt[0];

	return returnValue;
}

void upperHash_insert(upperHash *aUpperHash, sequence *aSequence)
{
	int slotNo;

	lowerHash *aLowerHash;

	aUpperHash->noOfEntry++;
	
	slotNo = upperHash_slotNo(aSequence);

	aLowerHash = aUpperHash->slot[slotNo];

	lowerHash_insert(aLowerHash, aSequence);
}




int upperHash_isEmpty(upperHash *aUpperHash)
{
	int returnValue;

	//int i;

	//int isEmpty;

	//returnValue = 1;
	//for (i = 0; i < aUpperHash->noOfSlot;i++)
	//{
	//	aLowerHash = aUpperHash->slot[i];

	//	isEmpty = lowerHash_isEmpty(aLowerHash);

	//	if (isEmpty == 0)
	//	{
	//		returnValue = 0;
	//		break;
	//	}
	//}

	if (aUpperHash->noOfEntry == 0)
	{
		returnValue = 1;
	}
	else
	{
		returnValue = 0;
	}


	return returnValue;
}


void lowerHash_free(lowerHash *aLowerHash)
{
	int i;

	if (aLowerHash->noOfSlot > 0)
	{
		for (i = 0; i < aLowerHash->noOfSlot; i++)
		{
			sequenceSet_free(aLowerHash->slot[i]);
		}
		free(aLowerHash->slot);
	}
	free(aLowerHash);
}


void upperHash_free(upperHash *aUpperHash)
{
	int i;

	if (aUpperHash->noOfSlot > 0)
	{
		for (i = 0; i < aUpperHash->noOfSlot; i++)
		{
			lowerHash_free(aUpperHash->slot[i]);
		}
		free(aUpperHash->slot);
	}
	free(aUpperHash);
}

upperHash *upperHash_clone(upperHash *aUpperHash)
{
	upperHash *returnValue;
	int i, j;

	lowerHash *aLowerHash;

	int upper_noOfSlot;
	int lower_noOfSlot;

	sequenceSet *aSequenceSet;

	sequenceSet *from_sequenceSet;

	lowerHash *from_lowerHash;

	lowerHash **from_upperSlot;


	upper_noOfSlot = aUpperHash->noOfSlot;
	lower_noOfSlot = aUpperHash->slot[0]->noOfSlot;

	returnValue = (upperHash *) malloc(sizeof(upperHash));
	memset(returnValue, 0, sizeof(upperHash));

	returnValue->noOfSlot = upper_noOfSlot;
	returnValue->slot = (lowerHash **) malloc(sizeof(lowerHash *)*upper_noOfSlot);


	from_upperSlot = aUpperHash->slot;

	for (i = 0; i < upper_noOfSlot; i++)
	{
		aLowerHash = (lowerHash *) malloc(sizeof(lowerHash));
		memset(aLowerHash, 0, sizeof(lowerHash));

		returnValue->slot[i] = aLowerHash;

		aLowerHash->noOfSlot = lower_noOfSlot;
		aLowerHash->slot = (sequenceSet **) malloc(sizeof(sequenceSet *)*lower_noOfSlot);

		from_lowerHash = from_upperSlot[i];
		
		for (j = 0; j < lower_noOfSlot; j++)
		{
			from_sequenceSet = from_lowerHash->slot[j];

			aSequenceSet = sequenceSet_clone(from_sequenceSet);
			aLowerHash->slot[j] = aSequenceSet;

		}
	}

	return returnValue;
}

void lowerHash_removeService(lowerHash *aLowerHash, int fromNo, int toNo)
{
	int noOfSlot;

	sequence *aSequence;
	sequenceSet *aSequenceSet;

	int slotNo;

	aSequence = sequence_new_noOfElt(2);
	aSequence->elt[0] = fromNo;
	aSequence->elt[1] = toNo;

	noOfSlot = aLowerHash->noOfSlot;
	slotNo = lowerHash_slotNo(aSequence, noOfSlot);

	aSequenceSet = aLowerHash->slot[slotNo];

	sequenceSet_remove(aSequenceSet, aSequence);

	sequence_free(aSequence);
}

void upperHash_removeService(upperHash *aUpperHash, int fromNo, int toNo)
{
	lowerHash *aLowerHash;

	aLowerHash = aUpperHash->slot[fromNo];
	aUpperHash->noOfEntry--;

	lowerHash_removeService(aLowerHash, fromNo, toNo);
}

void lowerHash_print(lowerHash *aLowerHash)
{
	int i;

	sequenceSet *aSeqSet;

	for (i = 0; i < aLowerHash->noOfSlot; i++)
	{
		printf("   Lower Slot No. %d\n", i);
		aSeqSet = aLowerHash->slot[i];

		sequenceSet_print(aSeqSet);
	}
}

void upperHash_print(upperHash *aUpperHash)
{
	int i;

	lowerHash *aLowerHash;

	for (i = 0; i < aUpperHash->noOfSlot; i++)
	{
		printf(" Upper Slot no. %d\n", i);
		aLowerHash = aUpperHash->slot[i];

		lowerHash_print(aLowerHash);
	}
}
