
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef WIN32
#include <sys/resource.h>
#include <sys/times.h>
#endif

#include "config.h"
#include "readGen.h"
#include "service.h"
#include "statistics.h"




#ifndef WIN32
void calculateExecutionTime(struct rusage *myTimeStart, struct rusage *myTimeEnd, float *userTime, float *sysTime)
{
	(*userTime) =
		((float) (myTimeEnd->ru_utime.tv_sec  - myTimeStart->ru_utime.tv_sec)) +
		((float) (myTimeEnd->ru_utime.tv_usec - myTimeStart->ru_utime.tv_usec)) * 1e-6;
	(*sysTime) =
		((float) (myTimeEnd->ru_stime.tv_sec  - myTimeStart->ru_stime.tv_sec)) +
		((float) (myTimeEnd->ru_stime.tv_usec - myTimeStart->ru_stime.tv_usec)) * 1e-6;
	
}
#endif

int countNoOfTupleWithCompetitiveService(data *aData, upperHash *competitiveServiceSet)
{
	int returnValue;
	int i, j;

	int isExist;

	sequence *aSequence;

	int noOfTuple;
	sequence *tuple;
	sequence *oneTuple;

	noOfTuple = aData->noOfTuple;
	tuple = aData->tuple;

	returnValue = 0;

	aSequence = sequence_new_noOfElt(2);
	aSequence->elt[0] = 0;
	aSequence->elt[1] = 0;

	for (i = 0; i < noOfTuple; i++)
	{
		oneTuple = &(tuple[i]);

		for (j = 0; j < oneTuple->noOfElt-1; j++)
		{
			aSequence->elt[0] = oneTuple->elt[j];

			aSequence->elt[1] = oneTuple->elt[j+1];

			isExist = upperHash_isExist(competitiveServiceSet, aSequence);
			if (isExist == 1)
			{
				returnValue++;
			}

		}
	}

	sequence_free(aSequence);

	return returnValue;
}

void polap_main(config *aConfig, data *aData, generalizedDimTable *genHierarchy, mySetSet *closeSetSet, int *ownerTable)
{
#ifndef WIN32
	float userTime_computeCompetitiveServiceSet, sysTime_computeCompetitiveServiceSet;
	float userTime_computeAllCompetitionAwareNode, sysTime_computeAllCompetitionAwareNode;
	float userTime_computeAllGroundSet, sysTime_computeAllGroundSet;
	float userTime_totalQuery, sysTime_totalQuery;
	float userTime_all, sysTime_all;
	float userTime_preprocess, sysTime_preprocess;
	float userTime_avgQuery, sysTime_avgQuery;

	struct rusage myTime_start;
	struct rusage myTime_computeCompetitiveServiceSet_end;
	struct rusage myTime_computeAllCompetitionAwareNode_end;
	struct rusage myTime_computeAllGroundSet_end;
	struct rusage myTime_query_end;

	char *timeFilename = "time.txt";
	FILE *timeFilePtr;
#endif

	char *debugFilename = "debug.txt";

	FILE *debugFilePtr; 

	statistics *aStat;

	int i, j;
	int noOfOlapOperations;

	sequence *prev_sequence;
	sequence *new_sequence;

	int operationNo;

	int upper_noOfSlot, lower_noOfSlot;

	upperHash *potentialServiceHash;
	mySetSet *allCompetitionAwareNode;

	mySetSet **allGroundSet;

	upperHash *prev_PCSS;
	upperHash *new_PCSS;

	upperHash *competitiveServiceSet;

	int prev_isCompetitive;
	int new_isCompetitive;

	double new_PCSS_size;

	int sizeAllService;
	int sizeCompetitiveService;
	int noOfTupleWithCompetitiveService;

	noOfOlapOperations = 6;

#ifndef WIN32
	getrusage(RUSAGE_SELF,&myTime_start);
#endif

	aStat = stat_new();

	debugFilePtr = (FILE *) fopen(debugFilename, "w");
	if (debugFilePtr == NULL)
	{
		printf("File %s cannot be opened!\n", debugFilename);
		exit(0);
	}

	// compute competitiveServiceSet
	printf("Computing competitive services...\n");
	upper_noOfSlot = aConfig->noOfValue;
	lower_noOfSlot = aConfig->noOfLowerSlot;
	potentialServiceHash = upperHash_new(upper_noOfSlot, lower_noOfSlot);
	service_extractAllPotentialCompetitiveSerivceFromData(aData, potentialServiceHash, closeSetSet, ownerTable);

	// counting
	sizeAllService = stat_calculateUpperHashSize(potentialServiceHash);

	// print 
	//printf("Printing potential competitive services..\n");
	//upperHash_print(potentialServiceHash);

	competitiveServiceSet = service_findAllCompetitiveService(aConfig, potentialServiceHash, closeSetSet, ownerTable);

	// counting
	sizeCompetitiveService = stat_calculateUpperHashSize(competitiveServiceSet);

	// counting
	printf("sizeAllService = %d\n", sizeAllService);
	printf("sizeCompetitiveService = %d\n", sizeCompetitiveService);

	// counting
	noOfTupleWithCompetitiveService = countNoOfTupleWithCompetitiveService(aData, competitiveServiceSet);
	printf("noOfTupleWithCompetitiveService = %d\n", noOfTupleWithCompetitiveService);

	// free potential
	upperHash_free(potentialServiceHash);

	// print 
	//printf("Printing competitive services..\n");
	//upperHash_print(competitiveServiceSet);

	//stat
	// calculate the competitiveServiceSet size
	aStat->size_competitiveService = stat_calculateUpperHashSize(competitiveServiceSet);

#ifndef WIN32
	getrusage(RUSAGE_SELF,&myTime_computeCompetitiveServiceSet_end);
#endif

	// compute allCompetitionAwareNode
	printf("Computing competition aware nodes....\n");
	allCompetitionAwareNode = findAllCompetitionAware(aConfig, genHierarchy, closeSetSet);

	// printing
	//printf("Printing competition aware nodes...\n");
	//mySetSet_print(allCompetitionAwareNode);

#ifndef WIN32
	getrusage(RUSAGE_SELF,&myTime_computeAllCompetitionAwareNode_end);
#endif

	// compute all groundset
	printf("Computing ground sets...\n");
	allGroundSet = generateReverseGForAllNode(genHierarchy, aConfig->noOfValue);

	// print allGroundSet
	//printAllGroundSet(allGroundSet, genHierarchy->noOfLevel);

#ifndef WIN32
	getrusage(RUSAGE_SELF,&myTime_computeAllGroundSet_end);
#endif

	// staring queries
	printf("Starting queries...\n");
	for (i = 0; i < aConfig->noOfBatch; i++)
	{
if (i == 2)
{
printf("");
}

		if (i%100 == 0)
		{
			printf("Processing %d-th batch...\n", i);
		}
		// randomly generate a sequence
		prev_sequence = generateRandomSequence(aConfig->seqLength, genHierarchy);

		//** temp
		//prev_sequence = sequence_new_noOfElt(2);
		//prev_sequence->elt[0] = 0;
		//prev_sequence->elt[1] = 1;
		//prev_sequence->elt[2] = 0;
		//prev_sequence->elt[3] = 0;

		//prev_sequence->level[0] = 0;
		//prev_sequence->level[1] = 0;
		//prev_sequence->level[2] = 0;
		//prev_sequence->level[3] = 0;

		//printf("~~~~~~~~~~~~~~ Sequence ~~~~~~~~~~~~~~\n");
		//sequence_print(prev_sequence);
		//printf("\n");
		
		prev_PCSS = generatePCSSFromSummary(prev_sequence, aConfig, allGroundSet, competitiveServiceSet, allCompetitionAwareNode);

		// printing
		//printf("Printing prev_PCSS (first)....\n");
		//upperHash_print(prev_PCSS);

		//stat
		new_PCSS_size = stat_calculateUpperHashSize(prev_PCSS);
		aStat->totalSize_PCSS += new_PCSS_size;

		prev_isCompetitive = determineWhetherQueryCompetitiveFreeFromPCSS(aConfig, competitiveServiceSet, prev_PCSS, closeSetSet, ownerTable);

		//stat
		aStat->totalNoOfQueries++;

		//stat
		if (prev_isCompetitive == 0)
		{
			aStat->noOfRestrictedQueries++;

			aStat->totalQueryLength += 1;
			continue;
		}

		for (j = 1; j < aConfig->noOfQueries; j++)
		{
			operationNo = rand()%6;
			//operationNo = 1;


			//printf("****** olap operation = %d (%d, %d)*********\n", operationNo, i, j);

//if ((i == 10) && (j == 3))
//{
//	printf("");
//}
//printf("~~~~~0\n");
			new_isCompetitive = determineWhetherQueryCompetitionFree(prev_isCompetitive, operationNo, genHierarchy, aConfig, &new_sequence, &new_PCSS, prev_sequence, prev_PCSS, allCompetitionAwareNode, allGroundSet, competitiveServiceSet, closeSetSet, ownerTable);

			//stat
			aStat->totalNoOfQueries++;

			//printf("~~~~~~~~~~~~~~ Sequence ~~~~~~~~~~~~~~\n");
			//sequence_print(new_sequence);
			//printf("\n");
//printf("~~~~~1\n");
			// printing
			//printf("Printing new_PCSS (%d)....\n", j);
			//upperHash_print(new_PCSS);
			//printf("Finish Printing new_PCSS (%d)....\n", j);

			//stat
			new_PCSS_size = stat_calculateUpperHashSize(new_PCSS);
			aStat->totalSize_PCSS += new_PCSS_size;
//printf("~~~~~2\n");
			//stat
			if (new_isCompetitive == 0)
			{
				aStat->noOfRestrictedQueries++;
				
			}

			// free prev_sequence
			// free prev_PCSS
			sequence_free(prev_sequence);
			upperHash_free(prev_PCSS);
//printf("~~~~~3\n");
			prev_isCompetitive = new_isCompetitive;
			prev_PCSS = new_PCSS;
			prev_sequence = new_sequence; 

			if (new_isCompetitive == 0)
			{
				aStat->totalQueryLength += j+1;
				break;
			}
//printf("~~~~~4\n");
		}

		if (new_isCompetitive == 1)
		{
			aStat->totalQueryLength += aConfig->noOfQueries;
		}

		sequence_free(new_sequence);
		upperHash_free(new_PCSS);
	}

#ifndef WIN32
	// end time 
	getrusage(RUSAGE_SELF,&myTime_query_end);
#endif

	aStat->percentOfRestrictedQueries = aStat->noOfRestrictedQueries/((double) aStat->totalNoOfQueries);
	aStat->avgSize_PCSS = aStat->totalSize_PCSS/((double) aStat->totalNoOfQueries);

	aStat->avgQueryLength = aStat->totalQueryLength/((double)aConfig->noOfBatch);
	
	// stat
	stat_print(debugFilePtr, aStat);
	
	printf("noOfRestrictedQueries      : %d\n", aStat->noOfRestrictedQueries);
	printf("totalNoOfQueries           : %d\n", aStat->totalNoOfQueries);
	printf("percentOfRestrictedQueries : %.10f\n", aStat->percentOfRestrictedQueries);
	printf("totalSize_PCSS             : %.10f\n", aStat->totalSize_PCSS);
	printf("avgSize_PCSS               : %.10f\n", aStat->avgSize_PCSS);
	printf("size_competitiveService    : %d\n", aStat->size_competitiveService);
	printf("totalQueryLength           : %d\n", aStat->totalQueryLength);
	printf("avgQueryLength             : %.10f\n", aStat->avgQueryLength);

	fclose(debugFilePtr);

#ifndef WIN32
 	// output execution time 
	calculateExecutionTime(&myTime_start, &myTime_computeCompetitiveServiceSet_end, &userTime_computeCompetitiveServiceSet, &sysTime_computeCompetitiveServiceSet);
	calculateExecutionTime(&myTime_computeCompetitiveServiceSet_end, &myTime_computeAllCompetitionAwareNode_end, &userTime_computeAllCompetitionAwareNode, &sysTime_computeAllCompetitionAwareNode);
	calculateExecutionTime(&myTime_computeAllCompetitionAwareNode_end, &myTime_computeAllGroundSet_end, &userTime_computeAllGroundSet, &sysTime_computeAllGroundSet);
	calculateExecutionTime(&myTime_computeAllGroundSet_end, &myTime_query_end, &userTime_totalQuery, &sysTime_totalQuery);
	calculateExecutionTime(&myTime_start, &myTime_computeAllGroundSet_end, &userTime_preprocess, &sysTime_preprocess);
	calculateExecutionTime(&myTime_start, &myTime_query_end, &userTime_all, &sysTime_all);

	userTime_avgQuery = userTime_totalQuery/(aConfig->noOfBatch*aConfig->noOfQueries);
	sysTime_avgQuery = sysTime_totalQuery/(aConfig->noOfBatch*aConfig->noOfQueries);

	timeFilePtr = (FILE *)fopen(timeFilename, "w");
	if (timeFilePtr == NULL)
	{
		printf("File %s cannot be opened!\n", timeFilename);
		exit(0);
	}

	fprintf(timeFilePtr, "%.10f %.10f\n", userTime_computeCompetitiveServiceSet, sysTime_computeCompetitiveServiceSet);
	fprintf(timeFilePtr, "%.10f %.10f\n", userTime_computeAllCompetitionAwareNode, sysTime_computeAllCompetitionAwareNode);
	fprintf(timeFilePtr, "%.10f %.10f\n", userTime_computeAllGroundSet, sysTime_computeAllGroundSet);
	fprintf(timeFilePtr, "%.10f %.10f\n", userTime_totalQuery, sysTime_totalQuery);
	fprintf(timeFilePtr, "%.10f %.10f\n", userTime_avgQuery, sysTime_avgQuery);
	fprintf(timeFilePtr, "%.10f %.10f\n", userTime_preprocess, sysTime_preprocess);
	fprintf(timeFilePtr, "%.10f %.10f\n", userTime_all, sysTime_all);

	printf("Time\n");
	printf("----\n");
	printf("computeCompetitiveSericeSet: %.10f %.10f\n", userTime_computeCompetitiveServiceSet, sysTime_computeCompetitiveServiceSet);
	printf("computeAllCompetitionAwareNode: %.10f %.10f\n", userTime_computeAllCompetitionAwareNode, sysTime_computeAllCompetitionAwareNode);
	printf("computeAllGroundSet: %.10f %.10f\n", userTime_computeAllGroundSet, sysTime_computeAllGroundSet);
	printf("totalQuery: %.10f %.10f\n", userTime_totalQuery, sysTime_totalQuery);
	printf("avgQuery: %.10f %.10f\n", userTime_avgQuery, sysTime_avgQuery);
	printf("preprocess: %.10f %.10f\n", userTime_preprocess, sysTime_preprocess);
	printf("all: %.10f %.10f\n", userTime_all, sysTime_all);

	fclose(timeFilePtr);

#endif

}

int main()
{
	char *configFilename = "conf.txt";

	config *aConfig;
	data *aData;
	generalizedDimTable *genHierarchy;
	mySetSet *closeSetSet;
	int *ownerTable;

	aConfig = config_readConfig(configFilename);

	aData = data_readData(aConfig->dataFilename, aConfig->noOfTuple);
	genHierarchy = gen_readGen(aConfig->hierarchyFilename);
	closeSetSet = close_readClose(aConfig->closeFilename, aConfig->noOfValue);
	ownerTable = owner_readOwner(aConfig->ownershipFilename, aConfig->noOfValue);

	if (aConfig->alg == 1)
	{
		polap_main(aConfig, aData, genHierarchy, closeSetSet, ownerTable);
	}
	else
	{
		// 
	}

	return 0;
}
