
#include "sequence.h"


sequence *sequence_new()
{
	sequence *returnValue;

	returnValue = (sequence *) malloc(sizeof(sequence));
	memset(returnValue, 0, sizeof(sequence));

	return returnValue;
}


sequence *sequence_new_noOfElt(int noOfElt)
{
	sequence *returnValue;

	returnValue = sequence_new();

	if (noOfElt > 0)
	{
		returnValue->noOfElt = noOfElt;
		returnValue->elt = (int *) malloc(sizeof(int)*noOfElt);
		returnValue->level = (int *) malloc(sizeof(int)*noOfElt);
		memset(returnValue->elt, 0, sizeof(int)*noOfElt);
		memset(returnValue->level, 0, sizeof(int)*noOfElt);
	}

	return returnValue;
}

void sequence_free(sequence *aSequence)
{	
	if (aSequence->noOfElt > 0)
	{
		free(aSequence->level);
		free(aSequence->elt);
	}
	free(aSequence);

}

// do not care for "level"
int sequence_isEqual(sequence *aSeq1, sequence *aSeq2)
{
	int returnValue;

	int i;

	int noOfElt1, noOfElt2;
	int *elt1;
	int *elt2;

	noOfElt1 = aSeq1->noOfElt;
	elt1 = aSeq1->elt;

	noOfElt2 = aSeq2->noOfElt;
	elt2 = aSeq2->elt;

	if (noOfElt1 != noOfElt2)
	{
		returnValue = 0;
	}
	else
	{
		returnValue = 1;
		for (i = 0; i < noOfElt1; i++)
		{
			if (elt1[i] != elt2[i])
			{
				returnValue = 0;
				break;
			}
		}
	}

	return returnValue;
}

sequenceSet *sequenceSet_new()
{
	sequenceSet *returnValue;

	returnValue = (sequenceSet *) malloc(sizeof(sequenceSet));
	memset(returnValue, 0, sizeof(sequenceSet));

	return returnValue;
}

sequenceSet *sequenceSet_new_noOfElt(int noOfElt)
{
	sequenceSet *returnValue;

	returnValue = sequenceSet_new();
	returnValue->noOfElt = noOfElt;
	returnValue->elt = (sequence **) malloc(sizeof(sequence *)*noOfElt);
	memset(returnValue->elt, 0, sizeof(sequence *)*noOfElt);

	return  returnValue;
}

void sequenceSet_insert(sequenceSet *aSequenceSet, sequence *aSequence)
{
	int noOfElt;
	sequence **elt;
	
	noOfElt = aSequenceSet->noOfElt;
	elt = aSequenceSet->elt;

	if (noOfElt == 0)
	{
		elt = (sequence **) malloc(sizeof(sequence *));
	}
	else
	{
		elt = (sequence **) realloc(elt, sizeof(sequence *)*(noOfElt+1));
	}
	elt[noOfElt] = aSequence;

	aSequenceSet->noOfElt++;
	aSequenceSet->elt = elt;
}



void sequenceSet_free(sequenceSet *aSequenceSet)
{
	int i;

	if (aSequenceSet->noOfElt > 0)
	{
		for (i = 0; i < aSequenceSet->noOfElt; i++)
		{
			sequence_free(aSequenceSet->elt[i]);
		}
		free(aSequenceSet->elt);
	}
	free(aSequenceSet);
}

int sequenceSet_isExist(sequenceSet *aSequenceSet, sequence *aSequence)
{
	int returnValue;

	int noOfElt;
	sequence **elt;

	int isEqual;

	sequence *tempSeq;

	int i;

	noOfElt = aSequenceSet->noOfElt;
	elt = aSequenceSet->elt;

	returnValue = 0;
	for (i = 0; i < noOfElt; i++)
	{
		tempSeq = elt[i];

		isEqual = sequence_isEqual(tempSeq, aSequence);
		if (isEqual == 1)
		{
			returnValue = 1;
			break;
		}
	}

	return returnValue;
}

void sequenceSet_insertNonRedundant(sequenceSet *aSequenceSet, sequence *aSequence)
{
	int noOfElt;
	sequence **elt;

	int isExist;
	
	noOfElt = aSequenceSet->noOfElt;
	elt = aSequenceSet->elt;

	isExist = sequenceSet_isExist(aSequenceSet, aSequence);

	if (isExist == 0)
	{

		if (noOfElt == 0)
		{
			elt = (sequence **) malloc(sizeof(sequence *));
		}
		else
		{
			elt = (sequence **) realloc(elt, sizeof(sequence *)*(noOfElt+1));
		}
		elt[noOfElt] = aSequence;

		aSequenceSet->noOfElt++;
		aSequenceSet->elt = elt;
	}
}

sequence *sequence_clone(sequence *aSequence)
{
	sequence *returnValue;

	int noOfElt;
	int *elt;

	elt = aSequence->elt;
	noOfElt = aSequence->noOfElt;

	returnValue = sequence_new_noOfElt(noOfElt);
	memcpy(returnValue->elt, aSequence->elt, sizeof(int)*noOfElt);

	return returnValue;
}

sequenceSet *sequenceSet_clone(sequenceSet *aSequenceSet)
{
	sequenceSet *returnValue;

	int noOfElt;
	sequence **elt;

	sequence *aSequence;

	int i;

	noOfElt = aSequenceSet->noOfElt;
	elt = aSequenceSet->elt;

	returnValue = sequenceSet_new_noOfElt(noOfElt);

	for (i = 0; i < noOfElt; i++)
	{
		aSequence = elt[i];
		returnValue->elt[i] = sequence_clone(aSequence);
	}


	return returnValue;
}

void sequenceSet_removeIndex(sequenceSet *aSequenceSet, int index)
{
	int noOfElt;
	sequence **elt;

	sequence **newElt;

	noOfElt = aSequenceSet->noOfElt;
	elt = aSequenceSet->elt;

	if (noOfElt > 1)
	{
		newElt = (sequence **) malloc(sizeof(sequence *)*(noOfElt-1));
	}
	else
	{
		newElt = NULL;
	}

	if (index > 0)
	{
		memcpy(newElt, elt, sizeof(sequence *)*index);
	}
	if (index < noOfElt - 1)
	{
		memcpy(&(newElt[index]), &(elt[index+1]), sizeof(sequence *)*(noOfElt -index -1));
	}

}

void sequenceSet_remove(sequenceSet *aSequenceSet, sequence *aSequence)
{
	int noOfElt;
	sequence **elt;

	int isEqual;

	sequence *tempSeq;

	int i;

	noOfElt = aSequenceSet->noOfElt;
	elt = aSequenceSet->elt;

	for (i = 0; i < noOfElt; i++)
	{
		tempSeq = elt[i];

		isEqual = sequence_isEqual(tempSeq, aSequence);
		if (isEqual == 1)
		{
			sequenceSet_removeIndex(aSequenceSet, i);
			break;
		}
	}
}

void sequence_print(sequence *aSequence)
{
	int noOfElt;
	int *elt;
	int *level;

	int i;

	noOfElt = aSequence->noOfElt;
	elt = aSequence->elt;
	level = aSequence->level;

	printf("(");
	for (i = 0; i < noOfElt; i++)
	{
		printf("%d-%d", level[i], elt[i]);
		if (i != noOfElt-1)
		{
			printf(" ");
		}
	}
	printf(") ");

}

void sequenceSet_print(sequenceSet *aSeqSet)
{
	int noOfElt;
	sequence **elt;
	int i;

	sequence *aSeq;
	
	noOfElt = aSeqSet->noOfElt;;
	elt = aSeqSet->elt;

	printf("           (noOfElt = %d): ", noOfElt);
	for (i = 0; i < noOfElt; i++)
	{
		aSeq = elt[i];
		sequence_print(aSeq);
	}
	printf("\n");
}

