
#include "set.h"

mySet *mySet_new()
{
	mySet *returnValue;

	returnValue = (mySet *) malloc(sizeof(mySet));
	memset(returnValue, 0, sizeof(mySet));

	return returnValue;
}

mySet *mySet_new_noOfElt(int noOfElt)
{
	mySet *returnValue;

	returnValue = mySet_new();

	if (noOfElt > 0)
	{
		returnValue->noOfElt = noOfElt;
		returnValue->elt = (int *) malloc(sizeof(int)*noOfElt);
		memset(returnValue->elt, 0, sizeof(int)*noOfElt);
	}

	return returnValue;
}

void mySet_free(mySet *aSet)
{
	if (aSet->noOfElt > 0)
	{
		free(aSet->elt);
	}
	free(aSet);
}


// assume aSet1 is sorted
// and aSet2 is sorted
int mySet_isEqual(mySet *aSet1, mySet *aSet2)
{
	int returnValue;

	int i;

	int noOfElt1, noOfElt2;
	int *elt1;
	int *elt2;

	noOfElt1 = aSet1->noOfElt;
	elt1 = aSet1->elt;

	noOfElt2 = aSet2->noOfElt;
	elt2 = aSet2->elt;

	if (noOfElt1 != noOfElt2)
	{
		returnValue = 0;
	}
	else
	{
		returnValue = 1;
		for (i = 0; i < noOfElt1; i++)
		{
			if (elt1[i] != elt2[i])
			{
				returnValue = 0;
				break;
			}
		}
	}

	return returnValue;
}

int mySet_isExist(mySet *aSet, int no)
{
	int returnValue;

	int noOfElt;
	int *elt;

	int i;

	elt = aSet->elt;
	noOfElt = aSet->noOfElt;

	returnValue = 0;
	for (i = 0; i < noOfElt; i++)
	{
		if (elt[i] == no)
		{
			returnValue = 1;
			break;
		}
	}

	return returnValue;
}

void mySet_insert(mySet *aSet, int eltNo)
{
	int noOfElt;
	int *elt;

	noOfElt = aSet->noOfElt;
	elt = aSet->elt;

	if (noOfElt == 0)
	{
		elt = (int *) malloc(sizeof(int));
	}
	else
	{
		elt = (int *) realloc(elt, sizeof(int)*(noOfElt+1));
	}
	elt[noOfElt] = eltNo;

	aSet->noOfElt++;
	aSet->elt = elt;
}

void mySet_insert_nonRedundance(mySet *aSet, int eltNo)
{
	int noOfElt;
	int *elt;

	int isExist;

	isExist = mySet_isExist(aSet, eltNo);
	if (isExist == 0)
	{
		noOfElt = aSet->noOfElt;
		elt = aSet->elt;

		
		if (noOfElt == 0)
		{
			elt = (int *) malloc(sizeof(int));
		}
		else
		{
			elt = (int *) realloc(elt, sizeof(int)*(noOfElt+1));
		}
		elt[noOfElt] = eltNo;

		aSet->noOfElt++;
		aSet->elt = elt;
	}
}


mySetSet *mySetSet_new()
{
	mySetSet *returnValue;

	returnValue = (mySetSet *) malloc(sizeof(mySetSet));
	memset(returnValue, 0, sizeof(mySetSet));

	return returnValue;
}

mySetSet *mySetSet_new_noOfElt(int noOfElt)
{
	mySetSet *returnValue;

	returnValue = mySetSet_new();
	returnValue->noOfElt = noOfElt;
	returnValue->elt = (mySet **) malloc(sizeof(mySet *)*noOfElt);
	memset(returnValue->elt, 0, sizeof(mySet *)*noOfElt);

	return  returnValue;
}

void mySetSet_free(mySetSet *aSetSet)
{
	int i;
	if (aSetSet->noOfElt > 0)
	{
		for (i = 0; i < aSetSet->noOfElt; i++)
		{
			mySet_free(aSetSet->elt[i]);
		}
		free(aSetSet->elt);
	}
	free(aSetSet);
}

void mySetSet_insert(mySetSet *aSetSet, mySet *aSet)
{
	int noOfElt;
	mySet **elt;
	
	noOfElt = aSetSet->noOfElt;
	elt = aSetSet->elt;

	if (noOfElt == 0)
	{
		elt = (mySet **) malloc(sizeof(mySet *));
	}
	else
	{
		elt = (mySet **) realloc(elt, sizeof(mySet *)*(noOfElt+1));
	}
	elt[noOfElt] = aSet;

	aSetSet->noOfElt++;
	aSetSet->elt = elt;
}

int mySetSet_isExist(mySetSet *aSetSet, mySet *aSet)
{
	int returnValue;

	int noOfElt;
	mySet **elt;

	int isEqual;

	mySet *tempSet;

	int i;

	noOfElt = aSetSet->noOfElt;
	elt = aSetSet->elt;

	returnValue = 0;
	for (i = 0; i < noOfElt; i++)
	{
		tempSet = elt[i];

		isEqual = mySet_isEqual(tempSet, aSet);
		if (isEqual == 1)
		{
			returnValue = 1;
			break;
		}
	}

	return returnValue;
}

mySet *mySet_minus(mySet *set1, mySet *set2)
{
	mySet *returnValue;

	int noOfElt1;
	int *elt1;

	int noOfElt2;
	int *elt2;

	int *isExist;

	int oneElt;
	int count;
	int noOfRemain;

	int i;


	noOfElt1 = set1->noOfElt;
	elt1 = set1->elt;

	noOfElt2 = set2->noOfElt;
	elt2 = set2->elt;

	isExist = (int *) malloc(sizeof(int)*noOfElt1);
	memset(isExist, 0, sizeof(int)*noOfElt1);

	noOfRemain = 0;
	for (i = 0; i < noOfElt1; i++)
	{
		oneElt = elt1[i];
		isExist[i] = mySet_isExist(set2, oneElt);
		if (isExist[i] == 1)
		{
			noOfRemain++;
		}
	}

	returnValue = mySet_new_noOfElt(noOfRemain);

	count = 0;
	for (i = 0; i < noOfElt1; i++)
	{
		if (isExist[i] == 1)
		{
			oneElt = elt1[i];
			returnValue->elt[count] = oneElt;
			count++;
		}
	}


	free(isExist);

	return returnValue;
}

void mySet_print(mySet *aSet)
{
	int i;

	for (i = 0; i < aSet->noOfElt; i++)
	{
		printf("%d ", aSet->elt[i]);
	}
}

void mySetSet_print(mySetSet *aSetSet)
{
	int noOfElt;
	mySet **elt;

	int i;

	noOfElt = aSetSet->noOfElt;
	elt = aSetSet->elt;

	for (i = 0; i < noOfElt; i++)
	{
		printf("   Entry %d\n", i);
		mySet_print(elt[i]);
		printf("\n");
	}
}

